@@ -18,6 +18,7 @@
 #include "config.h"
 #include <string.h>
 #include <ctype.h>
+#include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -280,11 +281,8 @@ int main(int argc, char *argv[])
     }
 
 #ifdef CONFIG_FILE
-    if (!server) {
+    if (!server)
 	server = match_config_file(qstring);
-	if (verb && server)
-	    printf(_("Using server %s.\n"), server);
-    }
 #endif
 
     if (!server)
@@ -1355,7 +1353,6 @@ char *convert_in6arpa(const char *s)
 	/* check that this is a valid digit for an IPv6 address */
 	if (!((*p >= '0' && *p <= '9') || (*p >= 'a' && *p <= 'f') ||
 	      (*p >= 'A' && *p <= 'F'))) {
-	    free(ip);
 	    ip[character] = '\0';
 	    return ip;
 	}
